package net.sf.sql2java.fetching.sql;

import java.sql.Connection;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.List;

import net.sf.sql2java.ApplicationContext;
import net.sf.sql2java.common.beans.DatabaseConnection;
import net.sf.sql2java.common.beans.SourceFilter;
import net.sf.sql2java.common.beans.datatype.SqlDataType;
import net.sf.sql2java.conversion.Convertor;
import net.sf.sql2java.conversion.KeywordGroup;
import net.sf.sql2java.data.holders.Column;
import net.sf.sql2java.data.holders.Database;
import net.sf.sql2java.data.holders.Index;
import net.sf.sql2java.data.holders.IndexColumn;
import net.sf.sql2java.data.holders.Schema;
import net.sf.sql2java.data.holders.Table;
import net.sf.sql2java.fetching.Fetcher;
import net.sf.sql2java.fetching.FetchingException;

public class SQLFetcher implements Fetcher {
	private SourceFilter filter;
	private DatabaseConnection connection;
	private Convertor convertor;
	
    private List<String> types = new ArrayList<String>();
	
// -- Constructors ------------------------------------------------------------
	
	public SQLFetcher() {}
	
// -- Fetching ----------------------------------------------------------------
	
	@Override
	public Database fetch() throws FetchingException {
		try {
			// -- create the convertor
			convertor = ApplicationContext.getInstance().getConvertor();
			
			// -- connect to the database
			connection.connect();
			
			// -- fetch the database
			Database database = fetchDatabase();
			
			// -- return the store
			return database;
		} catch (Exception se) {
			if (se instanceof FetchingException)
				throw (FetchingException) se;
			else
				throw new FetchingException(se);
		}
	}
	
	private Database fetchDatabase() throws FetchingException {
		try {
			// -- create a processable as database
			Database database = new Database(connection.getCatalog());
			
			// -- set the database properties
			database.setDriver(connection.getDriverClass());
			database.setUrl(connection.getUrl());
			database.setUsername(connection.getUsername());
			database.setPassword(connection.getPassword());
			
			// -- fetch the children
			fetchSchemas(database);
			
			// -- return the database
			return database;
		} catch (Exception e) {
			if (e instanceof FetchingException)
				throw (FetchingException) e;
			else
				throw new FetchingException(e);
		}
	}
	
	private void fetchSchemas(Database database) throws FetchingException {
		try {
	    	// -- get the connection
	        Connection conn = connection.getDataSource().getConnection();
	        
	        // -- get the tables for the given schema. If the schema is a dummy
	        // -- schema (name is "") we need to exclude it and just query without
	        // -- it
	        ResultSet rs = conn.getMetaData().getSchemas();
	
	        while (rs.next()) {
	        	// -- get the table name
	        	String name = rs.getString("TABLE_SCHEM");
	        	
	        	// -- continue if this table isn't a part of the filter
	        	if ((filter != null) && (! filter.match(database.getName(), name)))
	        		continue;
	
	        	// -- create the table processable
	        	Schema schema = database.createSchema(name);
	        	
	        	// -- fetch the children
				fetchTables(schema);
	        	
	        	// -- add the processable to the database
	        	database.saveSchema(schema);
	        }
	
	        // -- close the resultset and the connection
	        rs.close();
	        conn.close();
		} catch (Exception e) {
			if (e instanceof FetchingException)
				throw (FetchingException) e;
			else
				throw new FetchingException(e);
		}
	}
	
	private void fetchTables(Schema schema) throws FetchingException {
		try {
	    	// -- get the connection
	        Connection conn = connection.getDataSource().getConnection();
	        
	        // -- get the tables for the given schema. If the schema is a dummy
	        // -- schema (name is "") we need to exclude it and just query without
	        // -- it
	        ResultSet rs = conn.getMetaData().getTables(
	        		schema.getDatabase().getName(),
	                schema.getName(),
	                "%",
	                getTypesAsStringArray()
	        );
	
	        while (rs.next()) {
	        	// -- get the table name
	        	String name = rs.getString("TABLE_NAME");
	        	String type = rs.getString("TABLE_TYPE");
	        	
	        	// -- continue if this table isn't a part of the filter
	        	if ((filter != null) && 
	        		(! filter.match(schema.getDatabase().getName(), 
	        					    schema.getName(), 
	        					    name)))
	        		continue;
	
	        	// -- create the table processable
	        	Table table = schema.createTable(name);
	        	
	        	// -- set the parameters
	        	table.setTableType(type);
	        	
	        	// -- load the children
	        	fetchColumns(table);
	        	fetchIndexes(table);
	        	fetchPrimaryKeys(table);
	        	fetchRelations(table);
	        	
	        	// -- add the processable to the database
	        	schema.saveTable(table);
	        }
	
	        // -- close the resultset and the connection
	        rs.close();
	        conn.close();
		} catch (Exception e) {
			if (e instanceof FetchingException)
				throw (FetchingException) e;
			else
				throw new FetchingException(e);
		}
    }

    @SuppressWarnings("unchecked")
	private void fetchColumns(Table table) throws FetchingException {
    	try {
	        // -- create the connection object
	        Connection conn = connection.getDataSource().getConnection();
	
	        // -- get the columns from the database
	        ResultSet rs = conn.getMetaData().getColumns(
	        		table.getSchema().getDatabase().getName(),
	                table.getSchema().getName(),
	                table.getName(),
	                "%"
	        );
	
	        while (rs.next()) {
	        	String name = rs.getString("COLUMN_NAME");
	        	
	        	if ((filter != null) && (! filter.match(
	        					table.getSchema().getDatabase().getName(), 
	        					table.getSchema().getName(), 
	        					table.getName(), 
	        					name
        			))) continue;
	        	
	            // -- create the column object
	        	Column column = table.createColumn(name);
	
	            // -- set the properties of the column
	        	column.setDataType(SqlDataType.valueOf(rs.getInt("DATA_TYPE")));
	        	column.setSize(rs.getInt("COLUMN_SIZE"));
	        	column.setDecimalDigits(rs.getInt("DECIMAL_DIGITS"));
	        	column.setNumbersPrecedingRadix(rs.getInt("NUM_PREC_RADIX"));
	        	column.setNullable(rs.getInt("NULLABLE"));
	        	column.setRemarks(rs.getString("REMARKS"));
	        	column.setDefaultValue(rs.getString("COLUMN_DEF"));
	        	column.setOrdinalPosition(rs.getInt("ORDINAL_POSITION"));
	        	column.setReadable(true);
	        	column.setWritable(true);
	        	
	        	// -- add the processable to the store
	        	table.saveColumn(column);
	        }
	
	        // -- close the resultset and the connection
	        rs.close();
	        conn.close();
    	} catch (Exception e) {
    		if (e instanceof FetchingException)
				throw (FetchingException) e;
			else
				throw new FetchingException(e);
		}
    }

    private void fetchPrimaryKeys(Table table) throws FetchingException {
    	// -- continue if this index attributes do not match the filter
    	if ((filter != null) && 
    		(! filter.match(table.getSchema().getDatabase().getName(), 
    						table.getSchema().getName(), 
    						table.getName())))
    		return;
    	
    	try {
	        // -- create the connection
	        Connection conn = connection.getDataSource().getConnection();
	        
	        // -- get the primary keys from the database
	        ResultSet rs = conn.getMetaData().getPrimaryKeys(
	        		table.getSchema().getDatabase().getName(),
	                table.getSchema().getName(),
	                table.getName()
	        );
	
	        while (rs.next()) {
	        	String columnName = rs.getString("COLUMN_NAME");
	        	
	        	// -- get the column from the table
	        	Column column = table.getColumn(columnName);
	        	
	        	if (column == null)
	        		throw new FetchingException(
	        				String.format("No column found for key '%s'", columnName));
	        	
	        	// -- set the key sequence
	        	column.setKeySequence(rs.getShort("KEY_SEQ"));

	        	// -- persist the changed colun
	        	table.saveColumn(column);
	        }
	
	        // -- close the resultset and the connection
	        rs.close();
	        conn.close();
    	} catch (Exception e) {
    		if (e instanceof FetchingException)
				throw (FetchingException) e;
			else
				throw new FetchingException(e);
    	}
    }
	
    private void fetchIndexes(Table table) throws FetchingException {
    	// -- continue if this index attributes do not match the filter
    	if ((filter != null) && 
    		(! filter.match(table.getSchema().getDatabase().getName(), 
    						table.getSchema().getName(), 
    						table.getName())))
    		return;
    	
    	try {
	        // -- create the connection
	        Connection conn = connection.getDataSource().getConnection();
	        
	        // -- get the primary keys from the database
	        ResultSet rs = conn.getMetaData().getIndexInfo(
	        		table.getSchema().getDatabase().getName(), 
	        		table.getSchema().getName(), 
	        		table.getName(), 
	        		true, 
	        		true
	        );
	
	        while (rs.next()) {
	        	String columnName = rs.getString("COLUMN_NAME");
	        	String name = rs.getString("INDEX_NAME");
	        	
	        	// -- check if there already is an index with the given name
	        	Index index = null;
        		if (table.hasIndex(name)) {
        			index = table.getIndex(name);
        		} else {
        			index = table.createIndex(name);
        			
    	        	index.setUnique(! rs.getBoolean("NON_UNIQUE"));
    	        	index.setQualifier(convertor.translateTo(KeywordGroup.JAVA, rs.getString("INDEX_QUALIFIER")));
        		}

        		// -- get the column
	        	Column column = table.getColumn(columnName);
        		
        		// -- create the index column
        		IndexColumn indexColumn = new IndexColumn(column);
        		indexColumn.setSortSequence(rs.getString("ASC_OR_DESC"));
        		indexColumn.setFilter(rs.getString("FILTER_CONDITION"));
	        	
        		// -- add the indexColumn to the index
        		index.saveColumn(indexColumn);
        		
        		// -- persist the index changes
        		table.saveIndex(index);
	        }
	
	        // -- close the resultset and the connection
	        rs.close();
	        conn.close();
    	} catch (Exception e) {
    		if (e instanceof FetchingException)
				throw (FetchingException) e;
			else
				throw new FetchingException(e);
    	}
    }
    
    private void fetchRelations(Table table) throws FetchingException {
    	
    }
    
// -- Getters -----------------------------------------------------------------

	public DatabaseConnection getConnection() {
		return connection;
	}

    public List<String> getTypes() {
        return types;
    }

    public String[] getTypesAsStringArray() {
    	// -- create the result
        String[] result = new String[getTypes().size()];

        // -- convert the types list to an array
        result = getTypes().toArray(result);

        // -- return the result
        return result;
    }
	
// -- Setters -----------------------------------------------------------------
	
	public void setConnection(DatabaseConnection connection) {
		this.connection = connection;
	}
	
	public void setFilter(SourceFilter filter) {
		this.filter = filter;
	}
	
	public void setTypes(List<String> types) {
		this.types = types;
	}
}
