package net.sf.sql2java.generator;

import java.io.StringWriter;

import net.sf.sql2java.ApplicationContext;
import net.sf.sql2java.common.parameters.StringParser;
import net.sf.sql2java.common.parameters.handlers.NodeParameterHandler;
import net.sf.sql2java.common.store.PackStore;
import net.sf.sql2java.common.store.StoreException;
import net.sf.sql2java.conversion.Convertor;
import net.sf.sql2java.conversion.KeywordGroup;
import net.sf.sql2java.data.holders.Database;
import net.sf.sql2java.data.node.Nodelet;
import net.sf.sql2java.generator.events.CamelCaseStringReference;
import net.sf.sql2java.generator.events.ParameterStringReference;
import net.sf.sql2java.generator.scripting.Script;
import net.sf.sql2java.generator.scripting.ScriptEnvironment;
import net.sf.sql2java.generator.template.Template;
import net.sf.sql2java.generator.writer.CodeWriter;
import net.sf.sql2java.packages.PackageReference;
import net.sf.sql2java.packages.TemplatePack;
import net.sf.sql2java.utils.GeneratorUtils;

import org.apache.log4j.Logger;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;
import org.apache.velocity.app.event.EventCartridge;

public class Generator {
    public static Logger LOGGER = Logger.getLogger(Generator.class);
    
    private GeneratorTarget target;
    private CodeWriter writer;
    
    private GeneratorCycle cycle;
    
    private PackStore packStore;
    
    private Convertor convertor;

// -- Constructors ------------------------------------------------------------
    
    public Generator() { }
    
// -- Process -----------------------------------------------------------------
    
    /**
     * Start processing the templates
     *
     * This is the actual starting point. We'll call this funtion to begin
     * generating the code. But before we can do that, we'll have to check if 
     * the generator is initialized.
     *
     * This is not the only thing we need to do. We'll also need to add the
     * resources for the templates and the scripts.
     *
     * @param database The database metadata container. This object contains all
     *                  the schemas, tables, columns and other data.
     *
     * @throws Exception thrown when something bad happend when processing the
     *                   database of schemas.
     */
    public void generate(Database database) {
    	// -- create the cycle
    	cycle = new GeneratorCycle();
    	
    	// -- create the convertor
    	convertor = ApplicationContext.getInstance().getConvertor();
    	
    	// -- add all generator properties to the cycle
    	cycle.clearGeneratorProperties();
    	cycle.addGeneratorProperty("log", Logger.getLogger(Script.class));
    	cycle.addGeneratorProperty("util", new GeneratorUtils());
    	cycle.addGeneratorProperty("configuration", ApplicationContext.getInstance().getConfiguration());
    	cycle.addGeneratorProperty("convert", convertor);
    	cycle.addGeneratorProperty("target", target);
    	
    	// -- get the packStore and processorStore
    	packStore = ApplicationContext.getInstance().getPackStore();
    	
    	// -- first generate everything for the database
    	generateNode(database);
    }
    
    protected void generateNode(Nodelet node) {
    	LOGGER.info(String.format(
				"Starting generation of node '%s'", 
				node.toString()
		));
    	
    	// -- add the processable to the cycle
    	cycle.addProcessableProperty(node.getType(), node);
    	
    	// -- iterate all the packages
    	for (PackageReference refPack : packStore.getPacks().values())
    		generatePack(node, refPack.getPack());
    	
    	// -- generate the node's children
    	for (String childKey : node.keySet()) {
    		for (Nodelet childNode : node.get(childKey).values()) {
    			generateNode(childNode);
    		}
    	}
    	
    	// -- clear the processable
    	cycle.clearProcessableProperties();
    }
    
    protected void generatePack(Nodelet node, TemplatePack pack) {
    	// -- create the scripting environment
    	ScriptEnvironment scriptEnvironment = new ScriptEnvironment(pack);
    	
    	// -- add all package properties to the cycle
    	cycle.clearPackageProperties();
    	cycle.addPackageProperty("pack", pack);
    	cycle.addPackageProperty("script", scriptEnvironment);
    	
    	// -- get all the template from the pack
    	for (Template template : pack.getTemplates(node.getType()))
    		generateTemplate(node, template);
    	
    	// -- clear the package properties
    	cycle.clearPackageProperties();
    }
    
    
    protected void generateTemplate(Nodelet node, Template template) {
    	// -- create a String Parser for the current processable
    	StringParser tfp = new StringParser(new NodeParameterHandler(node));
    	
    	// -- create the target package
    	cycle.getTargetPackage().clear();
    	cycle.getTargetPackage().add(convertor.translateTo(KeywordGroup.JAVA, target.getPackage()));
    	cycle.getTargetPackage().add(convertor.translateTo(KeywordGroup.JAVA, template.getPackage()));
    	String targetPackage = tfp.parse(cycle.getTargetPackage().getPackage());
    	
    	// -- create the target file
    	cycle.getTargetFile().clear();
    	cycle.getTargetFile().add(convertor.translateTo(KeywordGroup.JAVA, target.getPath()));
    	cycle.getTargetFile().add(convertor.translateTo(KeywordGroup.JAVA, target.getPackage()));
    	cycle.getTargetFile().add(convertor.translateTo(KeywordGroup.JAVA, template.getPackage()));
    	String targetDirectory = tfp.parse(cycle.getTargetFile().getPath());
    	
    	cycle.getTargetFile().add(convertor.translateTo(KeywordGroup.JAVA, template.getFilename()));
    	cycle.getTargetFile().addRaw(convertor.translateTo(KeywordGroup.JAVA, template.getExtension()));
    	String targetFile = tfp.parse(cycle.getTargetFile().getPath()); 
    	
    	// -- add all the template properties to the cycle
    	cycle.clearTemplateProperties();
    	cycle.addTemplateProperty("template", template);
    	cycle.addTemplateProperty("targetPackage", targetPackage);
    	cycle.addTemplateProperty("targetDirectory", targetDirectory);
    	cycle.addTemplateProperty("targetFile", targetFile);
    	
    	try {
    		// -- parse all the properties inside the cycle
    		cycle.parseAll(node);
	    	
    		// -- generate the result for the given template
    		String result = generateContent(node, template);
	    		
    		// -- add the processable to the targetFile
    		// -- write the result to the file
    		writer.write(targetFile, result);
    	
    	} catch (StoreException se) {
    		LOGGER.error(se.getMessage(), se);
    	} catch (Exception e) {
    		LOGGER.error(e.getMessage(), e);
    	}
    	
    	// -- clear all the template properties
    	cycle.clearTemplateProperties();
    }
    
    protected String generateContent(Nodelet node, Template template) throws Exception {
    	// -- create the context
        VelocityContext vc = new VelocityContext(cycle.getProperties());
        
        // -- get the scripting environment
        ScriptEnvironment scriptEnvironment = cycle.getProperty("script");
        scriptEnvironment.addAllGlobals(cycle.getProperties());
        
        // -- create the event cartridge
        EventCartridge ec = new EventCartridge();
        
        // -- add the listeners to the cartridge
        ec.addEventHandler(new ParameterStringReference(node));
//        ec.addEventHandler(new CamelCaseStringReference());
        
        // -- append the cartridge to the context
        ec.attachToContext(vc);
        
        // -- initialize velocity
        Velocity.init();

        // -- create the writer object. We want to encapsulate everything into a 
        // -- nice little String, but velocity only takes Writers or Streams.
        // -- Time to get the StringWriter
        StringWriter sw = new StringWriter();
        
        // -- do the actual processing
        Velocity.evaluate(vc, sw, "LOG", template.getContent());

        // -- lets close the writer and get it's result
        String result = sw.toString();
        
        // -- return the result
        return result;
    }
    
    
// -- Getters -----------------------------------------------------------------
    
    public GeneratorTarget getTarget() {
        return target;
    }
    
    public CodeWriter getWriter() {
    	return writer;
    }
    
// -- Setters -----------------------------------------------------------------
    
    public void setTarget(GeneratorTarget target) {
        this.target = target;
    }
    
    public void setWriter(CodeWriter writer) {
    	this.writer = writer;
    }
}
