package net.sf.sql2java.generator;

import java.util.HashMap;
import java.util.Map;

import net.sf.sql2java.common.parameters.StringParser;
import net.sf.sql2java.common.parameters.handlers.NodeParameterHandler;
import net.sf.sql2java.common.resources.Resource;
import net.sf.sql2java.data.node.Nodelet;

public class GeneratorCycle {
	private Map<String, Object> generatorProperties = new HashMap<String, Object>();
	private Map<String, Object> processableProperties = new HashMap<String, Object>();
	private Map<String, Object> packageProperties = new HashMap<String, Object>();
	private Map<String, Object> templateProperties = new HashMap<String, Object>();
	
	private Resource targetPackage;
	private Resource targetFile;
	
// -- Constructors ------------------------------------------------------------
	
	public GeneratorCycle() {
		targetPackage = new Resource();
		targetFile = new Resource();
	}
	
// -- Property Handlers -------------------------------------------------------
	
	public void addGeneratorProperty(String key, Object value) {
		generatorProperties.put(key, value);
	}
	
	public void addProcessableProperty(String key, Object value) {
		processableProperties.put(key, value);
	}
	
	public void addPackageProperty(String key, Object value) {
		packageProperties.put(key, value);
	}
	
	public void addTemplateProperty(String key, Object value) {
		templateProperties.put(key, value);
	}
	
	/**
	 * Get a property.
	 * 
	 * We'll check for the given property key in the following sequence:
	 * <li>Template Properties</li> 
	 * <li>Package Properties</li>
	 * <li>Generator Properties</li>
	 * If no property with the given key is found, <code>null</code> will be returned
	 * @param <T>
	 * @param key
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public <T> T getProperty(String key) {
		if (templateProperties.containsKey(key))
			return (T) templateProperties.get(key);
		else
			if (processableProperties.containsKey(key))
				return (T) processableProperties.get(key);
			else
				if (packageProperties.containsKey(key))
					return (T) packageProperties.get(key);
				else
					if (generatorProperties.containsKey(key))
						return (T) generatorProperties.get(key);
					else
						return null;
	}
	
	public Map<String, Object> getProperties() {
		Map<String, Object> result = new HashMap<String, Object>();
		
		result.putAll(generatorProperties);
		result.putAll(processableProperties);
		result.putAll(packageProperties);
		result.putAll(templateProperties);
		
		return result;
	}
	
	public void parseAll(Nodelet node) {
		StringParser ps = new StringParser();
		ps.setHandler(new NodeParameterHandler(node));
		
		parseAll(ps, generatorProperties);
		parseAll(ps, processableProperties);
		parseAll(ps, packageProperties);
		parseAll(ps, templateProperties);
	}
	
	private void parseAll(StringParser ps, Map<String, Object> properties) {
		for (String key : properties.keySet()) {
			Object value = properties.get(key);
			
			if (value instanceof String)
				properties.put(key, ps.parse(value.toString()));
			else if (value instanceof Resource) {
				String raw = ((Resource) value).getRaw();
				
				((Resource) value).setRaw(ps.parse(raw));
				
				properties.put(key, value);
			}
				
		}
	}
	
	public void clearTemplateProperties() {
		templateProperties.clear();
		
		targetFile.clear();
		targetPackage.clear();
	}
	
	public void clearPackageProperties() {
		packageProperties.clear();
		
		clearTemplateProperties();
	}
	
	public void clearProcessableProperties() {
		processableProperties.clear();
		
		clearPackageProperties();
	}
	
	public void clearGeneratorProperties() {
		generatorProperties.clear();
		
		clearProcessableProperties();
	}
	
// -- Targets Resources -------------------------------------------------------
	
	public Resource getTargetPackage() {
		return targetPackage;
	}
	
	public Resource getTargetFile() {
		return targetFile;
	}
}
